@@ -4,9 +4,9 @@
     require DBI;
 
     @EXPORT = qw(); # Do NOT @EXPORT anything.
-    $VERSION = substr(q$Revision: 10.4 $, 9,-1);
+    $VERSION = sprintf("%d.%02d", q$Revision: 11.4 $ =~ /(\d+)\.(\d+)/o);
 
-#   $Id: NullP.pm,v 10.4 2001/05/29 23:25:55 timbo Exp $
+#   $Id: NullP.pm,v 11.4 2004/01/07 17:38:51 timbo Exp $
 #
 #   Copyright (c) 1994, Tim Bunce
 #
@@ -28,8 +28,8 @@
 	$drh;
     }
 
-    sub default_user {
-        return ('','');
+    sub CLONE {
+        undef $drh;
     }
 }
 
@@ -39,7 +39,6 @@
     use strict;
     # we use default (dummy) connect method
 
-    sub disconnect_all { }
     sub DESTROY { undef }
 }
 
